import asyncio
from typing import List

from realtime.streams import AudioStream, ByteStream, Stream, TextStream, VideoStream


def merge(input_queues: List[Stream]):
    output_queue = None
    if not all(isinstance(x, type(input_queues[0])) for x in input_queues):
        raise ValueError("All input queues must be of the same type")
    if isinstance(input_queues[0], AudioStream):
        output_queue = AudioStream()
    elif isinstance(input_queues[0], VideoStream):
        output_queue = VideoStream()
    elif isinstance(input_queues[0], TextStream):
        output_queue = TextStream()
    elif isinstance(input_queues[0], ByteStream):
        output_queue = ByteStream()
    else:
        raise ValueError(f"Invalid input queue type: {type(input_queues[0])}")

    for q in input_queues:

        async def get():
            while True:
                item = await q.get()
                output_queue.put_nowait(item)

        asyncio.create_task(get())
    return output_queue



import asyncio
import logging
import os
import time

import aiohttp

from realtime.plugins.base_plugin import Plugin
from realtime.streams import ByteStream, TextStream

logger = logging.getLogger(__name__)


class ElevenLabsTTS(Plugin):
    def __init__(
        self,
        *,
        api_key: str | None = None,
        voice_id: str = "pNInz6obpgDQGcFmaJgB",
        model: str = "eleven_turbo_v2",
        output_format: str = "pcm_16000",
        optimize_streaming_latency: int = 4,
        stream: bool = True,
    ):
        super().__init__()

        api_key = api_key or os.environ.get("ELEVEN_LABS_API_KEY")
        if api_key is None:
            raise ValueError("Eleven Labs API key is required")
        self._api_key = api_key
        self._voice_id = voice_id
        self._model = model
        self._output_format = output_format
        self._optimize_streaming_latency = optimize_streaming_latency

        self.output_queue = ByteStream()
        self._generating = False

        self._stream = stream

    # async def run_tts(self, sentence) -> Iterator[bytes]:
    #     Not using elevenlabs SDK since calling the API directly is 2x faster time to first chunk
    #     audio_stream = await self.aclient.generate(
    #         text=sentence + " ",
    #         voice=self._voice_id,
    #         model=self._model,
    #         stream=True,
    #         optimize_streaming_latency=4,
    #         output_format="pcm_16000",
    #     )
    #     async for chunk in audio_stream:
    #         if chunk:
    #             yield chunk

    async def run(self, input_queue: TextStream) -> ByteStream:
        self.input_queue = input_queue
        self._task = asyncio.create_task(self.synthesize_speech())
        return self.output_queue

    async def synthesize_speech(self):
        async with aiohttp.ClientSession() as self.session:
            while True:
                text_chunk = await self.input_queue.get()
                if text_chunk is None:
                    continue
                self._generating = True
                start_time = time.time()
                logger.info("Generating TTS %s", text_chunk)
                if self._stream:
                    url = f"https://api.elevenlabs.io/v1/text-to-speech/{self._voice_id}/stream"
                else:
                    url = f"https://api.elevenlabs.io/v1/text-to-speech/{self._voice_id}"
                payload = {"text": text_chunk, "model_id": self._model}
                querystring = {
                    "output_format": self._output_format,
                    "optimize_streaming_latency": self._optimize_streaming_latency,
                }
                headers = {
                    "xi-api-key": self._api_key,
                    "Content-Type": "application/json",
                }
                r = await self.session.post(url, json=payload, headers=headers, params=querystring)
                if r.status != 200:
                    logger.error("tts error %s", await r.text())
                    return

                logger.info("ElevenLabs TTS TTFB: %s", time.time() - start_time)
                if self._stream:
                    async for chunk in r.content:
                        if chunk:
                            self.output_queue.put_nowait(chunk)
                else:
                    self.output_queue.put_nowait(await r.read())

                self._generating = False

    async def close(self):
        await self.session.close()
        self._task.cancel()

    async def _interrupt(self):
        while True:
            user_speaking = await self.interrupt_queue.get()
            if self._generating and user_speaking:
                self._task.cancel()
                while not self.output_queue.empty():
                    self.output_queue.get_nowait()
                logger.info("Done cancelling TTS")
                self._generating = False
                self._task = asyncio.create_task(self.synthesize_speech())

    async def set_interrupt(self, interrupt_queue: asyncio.Queue):
        self.interrupt_queue = interrupt_queue
        self._interrupt_task = asyncio.create_task(self._interrupt())



import asyncio

from aiortc import MediaStreamTrack


class VideoOutputFrameProcessor(MediaStreamTrack):
    kind = "video"
    video_output_frame_q = asyncio.Queue()
    track = None

    def __init__(self, track=None):
        super().__init__()
        if track:
            VideoOutputFrameProcessor.track = track

    async def recv(self):
        frame = await VideoOutputFrameProcessor.video_output_frame_q.get()
        return frame

    async def put_frame(self, frame):
        await VideoOutputFrameProcessor.video_output_frame_q.put(frame)



import asyncio
import functools
import inspect
import logging
import os
import ssl
import time

import uvicorn

from realtime.streaming_endpoint.audio_frame_output import AudioOutputFrameProcessor
from realtime.streaming_endpoint.server import create_and_run_server
from realtime.streaming_endpoint.text_frame_output import TextFrameOutputProcessor
from realtime.streaming_endpoint.video_frame_output import VideoOutputFrameProcessor
from realtime.streams import AudioStream, TextStream, VideoStream

logger = logging.getLogger(__name__)


def streaming_endpoint():
    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            # Add the extra argument
            video_output_frame_processor = VideoOutputFrameProcessor()
            audio_output_frame_processor = AudioOutputFrameProcessor()
            text_output_processor = TextFrameOutputProcessor()

            try:
                audio_input_q = None
                video_input_q = None
                text_input_q = None
                instance = None

                signature = inspect.signature(func)
                parameters = signature.parameters
                for name, param in parameters.items():
                    if param.annotation == AudioStream:
                        audio_input_q = AudioStream()
                        kwargs[name] = audio_input_q
                    elif param.annotation == VideoStream:
                        video_input_q = VideoStream()
                        kwargs[name] = video_input_q
                    elif param.annotation == TextStream:
                        text_input_q = TextStream()
                        kwargs[name] = text_input_q
                instance = args[0]
                await instance.setup()
                output_streams = await func(*args, **kwargs)
                # Ensure output_streams is iterable
                if not isinstance(output_streams, (list, tuple)):
                    output_streams = (output_streams,)

                aq, vq, tq = None, None, None
                for s in output_streams:
                    if isinstance(s, AudioStream):
                        aq = s
                    elif isinstance(s, VideoStream):
                        vq = s
                    elif isinstance(s, TextStream):
                        tq = s

                input_audio = audio_input_q != None
                input_video = video_input_q != None
                output_audio = aq != None
                output_video = vq != None
                webrtc_app = create_and_run_server(input_audio, input_video, output_audio, output_video)

                HOSTNAME = "0.0.0.0"
                PORT = int(os.getenv("HTTP_PORT", 8080))
                ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
                ssl_context.load_cert_chain(os.environ["SSL_CERT_PATH"], keyfile=os.environ["SSL_KEY_PATH"])
                server = uvicorn.Server(
                    config=uvicorn.Config(
                        webrtc_app,
                        host=HOSTNAME,
                        port=PORT,
                        log_level="info",
                        ssl_keyfile=os.environ["SSL_KEY_PATH"],
                        ssl_certfile=os.environ["SSL_CERT_PATH"],
                    ),
                )
                asyncio.create_task(server.serve())

                async def audio_frame_callback():
                    try:
                        while not audio_output_frame_processor.track:
                            await asyncio.sleep(0.2)
                        audio_track = audio_output_frame_processor.track
                        while True:
                            frame = await audio_track.recv()
                            await audio_input_q.put(frame)
                    except Exception as e:
                        print("Error in audio_frame_callback: ", e)
                        raise asyncio.CancelledError

                async def video_frame_callback():
                    try:
                        while not video_output_frame_processor.track:
                            await asyncio.sleep(0.2)
                        video_track = video_output_frame_processor.track
                        while True:
                            frame = await video_track.recv()
                            await video_input_q.put(frame)
                    except Exception as e:
                        print("Error in video_frame_callback: ", e)
                        raise asyncio.CancelledError

                async def text_frame_callback():
                    try:
                        while not text_output_processor.track:
                            await asyncio.sleep(0.2)
                        while True:
                            text = await text_output_processor.recv()
                            print("text_frame_callback: ", text)
                            await text_input_q.put(text)
                    except Exception as e:
                        print("Error in text_frame_callback: ", e)
                        raise asyncio.CancelledError

                async def process_audio_output():
                    _start = None
                    data_time = 0.020
                    try:
                        while True:
                            if not aq:
                                break
                            if _start is None:
                                _start = time.time() + data_time
                            else:
                                wait = _start - time.time() - 0.005
                                _start += data_time
                                if wait > 0:
                                    await asyncio.sleep(wait)
                            try:
                                audio_frame = aq.get_nowait()
                            except asyncio.QueueEmpty:
                                continue
                            if audio_frame is None:
                                continue
                            await audio_output_frame_processor.put_frame(audio_frame)
                    except Exception as e:
                        print("Error in process_audio_output: ", e)
                        pass  # Ignore Empty exception

                async def process_video_output():
                    _start = None
                    data_time = 0.010
                    try:
                        while True:
                            if not vq:
                                break
                            if _start is None:
                                _start = time.time() + data_time
                            else:
                                wait = _start - time.time() - 0.005
                                _start += data_time
                                if wait > 0:
                                    await asyncio.sleep(wait)
                            try:
                                video_frame = vq.get_nowait()
                            except asyncio.QueueEmpty:
                                continue
                            if video_frame is None:
                                continue
                            await video_output_frame_processor.put_frame(video_frame)
                    except Exception as e:
                        print("Error in process_video_output: ", e)
                        pass  # Ignore Empty exception

                async def process_text_output():
                    try:
                        while True:
                            if not tq:
                                break
                            text_frame = await tq.get()
                            print("process_text_output: ", text_frame)
                            if text_frame is None:
                                continue
                            await text_output_processor.send(text_frame)
                    except Exception as e:
                        print("Error in process_text_output: ", e)
                        pass  # Ignore Empty exception

                tasks = [
                    asyncio.create_task(process_audio_output()),
                    asyncio.create_task(process_video_output()),
                    asyncio.create_task(process_text_output()),
                ]
                if audio_input_q:
                    tasks.append(asyncio.create_task(audio_frame_callback()))
                if video_input_q:
                    tasks.append(asyncio.create_task(video_frame_callback()))
                if text_input_q:
                    tasks.append(asyncio.create_task(text_frame_callback()))
                await asyncio.gather(*tasks)
            except Exception as e:
                VideoOutputFrameProcessor.track = None
                AudioOutputFrameProcessor.track = None
                TextFrameOutputProcessor.track = None
                if instance:
                    await instance.teardown()
                logging.info("Received exit, stopping bot")
                loop = asyncio.get_event_loop()
                tasks = asyncio.all_tasks(loop)
                for task in tasks:
                    task.cancel()
                    try:
                        loop.run_until_complete(task)
                    except asyncio.CancelledError:
                        logging.info("Task was cancelled")
                loop.close()

        return wrapper

    return decorator



from __future__ import annotations

import asyncio
import json
import logging
import os
from urllib.parse import urlencode

import aiohttp

from realtime.plugins.base_plugin import Plugin
from realtime.streams import TextStream

_KEEPALIVE_MSG: str = json.dumps({"type": "KeepAlive"})
_CLOSE_MSG: str = json.dumps({"type": "CloseStream"})


logger = logging.getLogger(__name__)


class DeepgramSTT(Plugin):
    def __init__(
        self,
        *,
        language="en-US",
        detect_language: bool = False,
        interim_results: bool = True,
        punctuate: bool = True,
        smart_format: bool = True,
        model="nova-2",
        api_key: str | None = None,
        sample_rate: int = 16000,
        num_channels: int = 1,
        min_silence_duration: int = 100,
        confidence_threshold=0.8,
    ) -> None:
        api_key = api_key or os.environ.get("DEEPGRAM_API_KEY")
        if api_key is None:
            raise ValueError("Deepgram API key is required")
        self._api_key = api_key
        self.language = language
        self.detect_language = detect_language
        self.interim_results = interim_results
        self.punctuate = punctuate
        self.smart_format = smart_format
        self.model = model
        self.min_silence_duration = min_silence_duration
        self.endpointing = min_silence_duration

        self._sample_rate = sample_rate
        self._num_channels = num_channels
        self._speaking = False
        self.confidence_threshold = confidence_threshold

        self._session = aiohttp.ClientSession()

        self._closed = False
        self.output_queue = TextStream()

    async def close(self):
        self.input_queue.put_nowait(_CLOSE_MSG)
        await asyncio.sleep(0.2)

        await self._session.close()
        self._task.cancel()

    async def run(self, input_queue) -> None:
        try:
            self.input_queue = input_queue
            self._task = asyncio.create_task(self._run_ws())
            return self.output_queue
        except Exception:
            logger.error("deepgram task failed")

    async def _run_ws(self) -> None:
        live_config = {
            "model": self.model,
            "punctuate": self.punctuate,
            "smart_format": self.smart_format,
            "encoding": "linear16",
            "sample_rate": self._sample_rate,
            "channels": self._num_channels,
            "endpointing": self.endpointing,
        }

        live_config["language"] = self.language

        headers = {"Authorization": f"Token {self._api_key}"}

        url = f"wss://api.deepgram.com/v1/listen?{urlencode(live_config).lower()}"
        ws = await self._session.ws_connect(url, headers=headers)

        async def keepalive_task():
            # if we want to keep the connection alive even if no audio is sent,
            # Deepgram expects a keepalive message.
            # https://developers.deepgram.com/reference/listen-live#stream-keepalive
            try:
                while True:
                    await ws.send_str(_KEEPALIVE_MSG)
                    await asyncio.sleep(5)
            except Exception:
                pass

        async def send_task():
            while True:
                data = await self.input_queue.get()

                if data == _CLOSE_MSG:
                    self._closed = True
                    await ws.send_str(data)
                    break

                bytes = data.to_ndarray().tobytes()
                await ws.send_bytes(bytes)

        async def recv_task():
            while True:
                msg = await ws.receive()
                if msg.type in (
                    aiohttp.WSMsgType.CLOSED,
                    aiohttp.WSMsgType.CLOSE,
                    aiohttp.WSMsgType.CLOSING,
                ):
                    if self._closed:
                        return

                    raise Exception("deepgram connection closed unexpectedly")

                if msg.type != aiohttp.WSMsgType.TEXT:
                    logger.error("unexpected deepgram message type %s", msg.type)
                    continue

                try:
                    data = json.loads(msg.data)
                    if not "is_final" in data:
                        break
                    is_final = data["is_final"]
                    top_choice = data["channel"]["alternatives"][0]
                    confidence = top_choice["confidence"]
                    if top_choice["transcript"] and confidence > self.confidence_threshold and is_final:
                        logger.info("deepgram transcript: %s", top_choice["transcript"])
                        await self.output_queue.put(top_choice["transcript"])
                except Exception as e:
                    logger.error("failed to process deepgram message %s", e)

        await asyncio.gather(send_task(), recv_task(), keepalive_task())



from typing import Type


def App():
    def wrapper(user_cls: Type):
        new_class = RealtimeApp(user_cls=user_cls)
        return new_class

    return wrapper


class RealtimeApp:
    def __init__(self, user_cls):
        self._user_cls = user_cls
        # self._realtime_functions = RealtimeFunction.get_realtime_functions_from_class(user_cls=_user_cls)
        self._user_cls_instance = None

    def __call__(self, *args, **kwargs):
        if self._user_cls_instance:
            raise Exception("This object is not callable.")
        self._user_cls_instance = self._user_cls(*args, **kwargs)
        return self._user_cls_instance.run()

    def __getattr__(self, k):
        if self._user_cls_instance:
            return getattr(self._user_cls_instance, k)
        else:
            raise AttributeError(k)



from typing import Any, Callable, Dict, Type


class RealtimeFunction:
    raw_f: Callable[..., Any]

    def __init__(
        self,
        raw_f: Callable[..., Any],
    ):
        self.raw_f = raw_f

    @classmethod
    def get_realtime_functions_from_class(cls, user_cls: Type):
        realtime_functions: Dict[str, RealtimeFunction] = {}
        for parent_cls in user_cls.mro():
            if parent_cls is not object:
                for k, v in parent_cls.__dict__.items():
                    if isinstance(v, RealtimeFunction):
                        realtime_functions[k] = v

        return realtime_functions



import asyncio
import logging
import os
import uuid
from contextlib import asynccontextmanager
from typing import Dict

from aiortc import RTCPeerConnection, RTCSessionDescription
from aiortc.contrib.media import MediaRelay
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from realtime.streaming_endpoint.audio_frame_output import AudioOutputFrameProcessor
from realtime.streaming_endpoint.text_frame_output import TextFrameOutputProcessor
from realtime.streaming_endpoint.video_frame_output import VideoOutputFrameProcessor

ROOT = os.path.dirname(__file__)

logger = logging.getLogger("pc")
logger.setLevel(logging.DEBUG)
pcs = asyncio.Queue()
relay = MediaRelay()


def offer(input_audio, input_video, output_audio, output_video):
    async def handshake(params: Dict[str, str]):
        offer = RTCSessionDescription(sdp=params["sdp"], type=params["type"])

        pc = RTCPeerConnection()
        pc_id = "PeerConnection(%s)" % uuid.uuid4()
        await pcs.put(pc)

        def log_info(msg, *args):
            logger.info(pc_id + " " + msg, *args)

        # log_info("Created for %s", params["remote"])

        @pc.on("datachannel")
        def on_datachannel(channel):
            text_output_processor = TextFrameOutputProcessor(channel)

            @channel.on("message")
            def on_message(message):
                text_output_processor.put_text(message)

        @pc.on("connectionstatechange")
        async def on_connectionstatechange():
            log_info("Connection state is %s", pc.connectionState)
            if pc.connectionState == "failed":
                await pc.close()
                pcs.discard(pc)

        # For some unknown reason, making this funciton async breaks aiortc
        @pc.on("track")
        def on_track(track):
            log_info("Track %s received", track.kind)

            # avoid using relay.subscribe since it doesn't close properly when connection is stopped
            if track.kind == "audio" and input_audio:
                AudioOutputFrameProcessor.track = track
            elif track.kind == "video" and input_video:
                VideoOutputFrameProcessor.track = track

        # handle offer
        await pc.setRemoteDescription(offer)
        if output_audio:
            pc.addTrack(AudioOutputFrameProcessor())
        if output_video:
            pc.addTrack(VideoOutputFrameProcessor())

        # send answer
        answer = await pc.createAnswer()
        await pc.setLocalDescription(answer)
        return {"sdp": pc.localDescription.sdp, "type": pc.localDescription.type}

    return handshake


@asynccontextmanager
async def on_shutdown(app: FastAPI):
    yield
    # close peer connections
    # coros = [pc.close() for pc in pcs]
    # await asyncio.gather(*coros)
    # pcs.clear()


def create_and_run_server(input_audio, input_video, output_audio, output_video):
    webrtc_app = FastAPI(lifespan=on_shutdown)
    webrtc_app.add_middleware(CORSMiddleware, allow_origins=["*"], allow_methods=["*"], allow_headers=["*"])
    webrtc_app.add_api_route("/offer", offer(input_audio, input_video, output_audio, output_video), methods=["POST"])
    return webrtc_app



import asyncio
import functools
import inspect
import logging
import os
import ssl
import time

import uvicorn

from fastapi.middleware.cors import CORSMiddleware
from fastapi import FastAPI

logger = logging.getLogger(__name__)


def web_endpoint(method: str = "POST", path: str = "/"):
    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            webrtc_app = FastAPI()
            webrtc_app.add_middleware(CORSMiddleware, allow_origins=["*"], allow_methods=["*"], allow_headers=["*"])
            webrtc_app.add_api_route(path, func, methods=[method])

            HOSTNAME = "0.0.0.0"
            PORT = int(os.getenv("HTTP_PORT", 8080))
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            ssl_context.load_cert_chain(os.environ["SSL_CERT_PATH"], keyfile=os.environ["SSL_KEY_PATH"])
            server = uvicorn.Server(
                config=uvicorn.Config(
                    webrtc_app,
                    host=HOSTNAME,
                    port=PORT,
                    log_level="info",
                    ssl_keyfile=os.environ["SSL_KEY_PATH"],
                    ssl_certfile=os.environ["SSL_CERT_PATH"],
                )
            )
            await server.serve()

        return wrapper

    return decorator



import asyncio

from realtime.plugins.base_plugin import Plugin
from realtime.streams import TextStream

SENTENCE_ENDINGS = [".", "!", "?", "\n"]


class TokenAggregator(Plugin):
    def __init__(self):
        super().__init__()
        self.output_queue = TextStream()
        self.buffer = ""

    async def run(self, input_queue: asyncio.Queue) -> asyncio.Queue:
        self.input_queue = input_queue
        self._task = asyncio.create_task(self._aggregate_tokens())
        return self.output_queue

    async def _aggregate_tokens(self):
        while True:
            token = await self.input_queue.get()
            if not token:
                continue
            self.buffer += token
            for ending in SENTENCE_ENDINGS:
                i = self.buffer.rfind(ending)
                if i != -1 and len(self.buffer[: i + 1]) >= 10:
                    await self.output_queue.put(self.buffer[: i + 1])
                    self.buffer = self.buffer[i + 1 :]
                    break

    async def close(self):
        self._task.cancel()

    async def _interrupt(self):
        while True:
            user_speaking = await self.interrupt_queue.get()
            if user_speaking and not self.output_queue.empty():
                self.buffer = ""
                self._task.cancel()
                while not self.output_queue.empty():
                    self.output_queue.get_nowait()
                print("Done cancelling Token Aggregator")
                self._task = asyncio.create_task(self._aggregate_tokens())

    async def set_interrupt(self, interrupt_queue: asyncio.Queue):
        self.interrupt_queue = interrupt_queue
        self._interrupt_task = asyncio.create_task(self._interrupt())



import asyncio


class TextFrameOutputProcessor:
    kind = "text"
    track = None
    text_received_q = asyncio.Queue()

    def __init__(self, track=None):
        super().__init__()
        if track:
            TextFrameOutputProcessor.track = track

    async def send(self, text):
        TextFrameOutputProcessor.track.send(text)

    def put_text(self, text):
        TextFrameOutputProcessor.text_received_q.put_nowait(text)

    async def recv(self):
        return await TextFrameOutputProcessor.text_received_q.get()



from concurrent.futures import ThreadPoolExecutor
import faulthandler
import json

faulthandler.enable()


import asyncio
import logging
import os
import time

import aiohttp

from realtime.plugins.base_plugin import Plugin
from realtime.streams import ByteStream, TextStream
import azure.cognitiveservices.speech as speechsdk

logger = logging.getLogger(__name__)


viseme_id_to_mouth_shapes = {
    0: "X",
    1: "D",
    2: "D",
    3: "F",
    4: "C",
    5: "C",
    6: "B",
    7: "F",
    8: "E",
    9: "E",
    10: "F",
    11: "D",
    12: "C",
    13: "F",
    14: "C",
    15: "B",
    16: "F",
    17: "H",
    18: "G",
    19: "H",
    20: "H",
    21: "A",
}


class AzureTTS(Plugin):
    def __init__(
        self,
        *,
        api_key: str | None = None,
        voice_id: str = "en-US-AvaMultilingualNeural",
        output_format: str = "pcm_16000",
        optimize_streaming_latency: int = 4,
        stream: bool = True,
        azure_speech_region: str | None = None,
    ):
        super().__init__()

        azure_speech_key = api_key or os.getenv("AZURE_SPEECH_KEY")
        azure_speech_region = azure_speech_region or os.getenv("AZURE_SPEECH_REGION")
        if not azure_speech_key:
            raise ValueError("Please set AZURE_SPEECH_KEY environment variable or pass it as a parameter")
        if not azure_speech_region:
            raise ValueError("Please set AZURE_SPEECH_REGION environment variable or pass it as a parameter")
        self._api_key = azure_speech_key
        self._voice_id = voice_id
        self._output_format = output_format
        self._optimize_streaming_latency = optimize_streaming_latency

        self.output_queue = ByteStream()
        self.viseme_stream = TextStream()
        self._viseme_data = {"mouthCues": []}
        self._generating = False
        self.thread_pool_executor = ThreadPoolExecutor(max_workers=1)

        self._stream = stream
        self._speech_config = speechsdk.SpeechConfig(subscription=azure_speech_key, region=azure_speech_region)
        self._speech_config.set_speech_synthesis_output_format(
            speechsdk.SpeechSynthesisOutputFormat.Raw16Khz16BitMonoPcm
        )
        self._speech_config.speech_synthesis_voice_name = voice_id
        try:
            self._speech_synthesizer = speechsdk.SpeechSynthesizer(speech_config=self._speech_config, audio_config=None)
            self._speech_synthesizer.viseme_received.connect(self.viseme_received_cb)
            # connection = speechsdk.Connection.from_speech_synthesizer(self._speech_synthesizer)
            # connection.open(True)
        except Exception as e:
            logger.debug(f"Error while connecting to Azure synthesizer: {e}")

    def viseme_received_cb(self, evt: speechsdk.SessionEventArgs):
        if len(self._viseme_data["mouthCues"]) > 0:
            self._viseme_data["mouthCues"][-1]["end"] = (evt.audio_offset) / 10000000.0
        self._viseme_data["mouthCues"].append(
            {
                "value": viseme_id_to_mouth_shapes[evt.viseme_id],
                "start": (evt.audio_offset) / 10000000.0 if len(self._viseme_data["mouthCues"]) > 0 else 0.0,
                "end": (evt.audio_offset + 30000.0) / 10000000.0,
            }
        )

    async def run(self, input_queue: TextStream) -> ByteStream:
        self.input_queue = input_queue
        self._task = asyncio.create_task(self.synthesize_speech())
        return self.output_queue, self.viseme_stream

    async def synthesize_speech(self):
        while True:
            text_chunk = await self.input_queue.get()
            if text_chunk is None:
                continue
            self._generating = True
            start_time = time.time()
            logger.info("Generating TTS %s", text_chunk)

            if self._stream:
                result = await asyncio.get_event_loop().run_in_executor(
                    self.thread_pool_executor,
                    lambda: self._speech_synthesizer.start_speaking_text_async(text_chunk).get(),
                )
                audio_data_stream = speechsdk.AudioDataStream(result)
                audio_buffer = bytes(4000)
                filled_size = await asyncio.get_event_loop().run_in_executor(
                    self.thread_pool_executor,
                    lambda: audio_data_stream.read_data(audio_buffer),
                )
                logger.info("Azure TTS TTFB: %s", time.time() - start_time)
                while filled_size > 0:
                    await self.output_queue.put(audio_buffer)
                    filled_size = await asyncio.get_event_loop().run_in_executor(
                        self.thread_pool_executor,
                        lambda: audio_data_stream.read_data(audio_buffer),
                    )
            else:
                result = await asyncio.get_event_loop().run_in_executor(
                    self.thread_pool_executor,
                    lambda: self._speech_synthesizer.speak_text_async(text_chunk).get(),
                )
                audio_data = result.audio_data
                logger.info("Azure TTS TTFB: %s", time.time() - start_time)
                await self.output_queue.put(audio_data)
                await self.viseme_stream.put(json.dumps(self._viseme_data))
                self._viseme_data["mouthCues"] = []

            self._generating = False

    async def close(self):
        self._task.cancel()

    async def _interrupt(self):
        while True:
            user_speaking = await self.interrupt_queue.get()
            if self._generating and user_speaking:
                self._task.cancel()
                while not self.output_queue.empty():
                    self.output_queue.get_nowait()
                logger.info("Done cancelling TTS")
                self._generating = False
                self._task = asyncio.create_task(self.synthesize_speech())

    async def set_interrupt(self, interrupt_queue: asyncio.Queue):
        self.interrupt_queue = interrupt_queue
        self._interrupt_task = asyncio.create_task(self._interrupt())
